import System.Environment

import Control.Monad.State
import Control.DeepSeq
import Control.Parallel.Strategies

import qualified Data.List as L
import qualified Data.Set as S

------------------------------
-- Monadic assignment trail --

data Choice = Choice { literal :: Int, visited :: Bool }
            | Implication { literal :: Int }
            deriving (Eq, Show)

instance NFData Choice where
    rnf (Choice i v) = rnf i `seq` rnf v
    rnf (Implication i) = rnf i

type Trail = [Choice]
type Model = S.Set Int
type Solution = (Trail, Model, Int)

pop :: State Solution (Maybe Choice)
pop = state pop'
    where pop' (x@(Choice i _):xs, m, d) = (Just x, (xs, S.delete i m, d - 1))
          pop' (x@(Implication i):xs, m, d) = (Just x, (xs, S.delete i m, d))
          pop' s@([], _, _) = (Nothing, s)

push :: Choice -> State Solution Bool
push x@(Choice i _) = state (\(xs, m, d) -> (True, (x:xs, S.insert i m, d + 1)))
push x@(Implication i) = state (\(xs, m, d) -> (True, (x:xs, S.insert i m, d)))

model :: State Solution Model
model = (\(_, m, _) -> m) <$> get

depth :: State Solution Int
depth = (\(_, _, d) -> d) <$> get

printSolution :: Solution -> IO ()
printSolution (_, m, _) = putStrLn $ "Sat\n" ++ (unwords $ map show $ S.elems m)

----------------------
-- Unit propagation --

type Clause = [Int]
data Status = Safe | Conflict | Implies [Int] deriving (Eq, Show)

litStatus :: S.Set Int -> Int -> Status
litStatus m i
    | S.member i m          = Safe
    | S.member (negate i) m = Conflict
    | otherwise             = Implies [i]

clauseStatus :: S.Set Int -> Clause -> Status
clauseStatus _ []     = Conflict
clauseStatus m (i:is) = case litStatus m i of
   Safe     -> Safe
   Conflict -> continue
   imp      -> if continue == Conflict then imp else Safe
   where continue = clauseStatus m is

status :: S.Set Int -> [Clause] -> Status
status _ []     = Safe
status m (c:cs) = case clauseStatus m c of
    Safe        -> continue
    Conflict    -> Conflict
    Implies [i] -> case continue of
        Implies is -> if negate i `elem` is then Conflict else Implies (i:is)
        Safe       -> Implies [i]
        Conflict   -> Conflict
    _ -> error "bad Implies"
    where continue = status m cs

unitPropagate :: [Clause] -> State Solution Bool
unitPropagate cs = do
    m <- model
    case status m cs of
        Safe       -> return False
        Conflict   -> return True
        Implies is -> do mapM_ (push . Implication) $ L.nub is
                         unitPropagate cs

-------------------------
-- Core DPLL algorithm --

backtrack :: State Solution Bool
backtrack = do
    choice <- pop
    case choice of
        Just (Choice i False) -> push $ Choice (negate i) True
        Just _                -> backtrack
        Nothing               -> return False

propagate :: [Clause] -> State Solution Bool
propagate cs = do
    conflict <- unitPropagate cs
    if conflict
        then do
            continue <- backtrack
            if continue then propagate cs else return False
        else dpll cs

next :: [Clause] -> State Solution (Maybe Int)
next cs = do
    m <- model
    let assigned i = S.member i m || S.member (negate i) m
    return $ L.find (not . assigned) (concat cs)

bothBranches :: Model -> Int -> [Clause] -> Int -> Eval (State Solution Bool)
bothBranches m d cs i = do
    let forwSt = ([Choice i True], S.insert i m, d + 1)
        forkSt = ([Choice (negate i) True], S.insert (negate i) m, d + 1)
    fork <- rpar $ force $ runState (propagate cs) forkSt
    forw <- rseq $ runState (propagate cs) forwSt
    case forw of
        (True, sol) -> return $ state (\(_, _, _) -> (True, sol))
        (False, _) -> do
            _ <- rseq fork
            case fork of
                (True, sol) -> return $ state (\(_, _, _) -> (True, sol))
                (False, _) -> return $ return False

dpll :: [Clause] -> State Solution Bool
dpll cs = do
    n <- next cs
    case n of
        Just i -> do
            m <- model
            d <- depth
            runEval $ bothBranches m d cs i
        Nothing -> return True

-------------------------
-- Preprocessing steps --

unfalsifiable :: Clause -> Bool
unfalsifiable c = (length $ L.nub $ map abs c) /= (length c)

sanitize :: [Clause] -> [Clause]
sanitize cs = L.sortOn length $ filter (not . unfalsifiable) $ L.nub cs

-------------
-- Parsing --

cnfToClauses :: [String] -> [Clause]
cnfToClauses cnf = map (L.nub . init . map read . words) cnf

parseCnf :: String -> IO [Clause]
parseCnf f = cnfToClauses . tail . lines <$> readFile f

main :: IO ()
main = do
    args <- getArgs
    case args of
        [f] -> do
            cs <- parseCnf f
            case runState (dpll $ sanitize cs) ([], S.empty, 0) of
                (False, _) -> putStrLn "Unsat"
                (True, sol) -> printSolution sol
        _ -> putStrLn "Usage: ./sat-parallel file.cnf"
